var watch = require("watch"),
	less = require("less"),
	colors = require("colors"),
	growl = require("./growl"),
	sys = require("sys"),
	util = require("util"),
	fs = require("fs");

var compileTo = "",
	cwd = "",
	buffer = "",
	viewExt = "",
	lessFolders = [],
	viewFolders = [],
	lessFiles = [],
	viewFiles = [],
	compileError = false,
	fileDelimiter = "//::file::";


// String helpers
String.prototype.__defineGetter__('ext',function() { return this.split(".")[this.split(".").length-1]; });
String.prototype.replaceExt = function(value) {	return this.replace("."+this.ext,"."+value); };


var notify =
{
	added : function(message)
	{
		sys.puts("\n"+message.green);
		sys.puts("Recompiling CSS.");
		notify.growl(message, "Added");
	},
	removed : function(message)
	{
		sys.puts("\n"+message.red);
		sys.puts("Recompiling CSS.");
		notify.growl(message, "Removed");
	},
	changed : function(message)
	{
		sys.puts("\n"+message.yellow);
		sys.puts("Recompiling CSS.");
		notify.growl(message, "Updated");
	},
	error : function(message) {
		sys.puts("\n"+message.red);
		notify.growl(message, "Error!");
	},
	success : function(message) {
		sys.puts("\n"+message.green);
		notify.growl(message, "Success");
	},
	growl : function(message, title) {
		growl.notify(message, {title:title,image:__dirname+"/icon.icns"});
	}
};

lessFiles.add = function(file,content) {
	this.push({
		file: file,
		content: content ? content : readLessContent(file)
	});
};

lessFiles.find = function(file) {
	return this.filter(function(f){
		if (f.file == file)
			return true;
	})[0];
};

lessFiles.update = function(file) {
	var f = this.find(file);

	if (f)
		f.content = readLessContent(file);
	else
		notify.error("Couldn't find file to update.");
};

lessFiles.remove = function(file) {
	var f = this.find(file);

	if (f)
		this.splice(this.indexOf(f),1);
	else
		notify.error("Couldn't find file to remove.");
};


// Get less file content
function readLessContent(file)
{
	// Mark beginning of file for error reporting
	var content = fileDelimiter + file.replace(cwd,"") + "\n";

	// Add file content to buffer
	content += fs.readFileSync(file,["utf-8"]);

	return content;
}


// Do the deed
function compileLess()
{
	buffer = "";
	lessFiles.forEach(function(file){
		buffer += file.content;
	});

	var parser = new(less.Parser);

	try {

		parser.parse(buffer,function(err, tree)
		{
			if (err) {

				outputLessError(err, buffer);

			} else {

				var css = tree.toCSS();

				fs.writeFile(cwd+compileTo, css, function(err)
				{
					if (err) notify.error("Could not write CSS file.\n"+err.message);
				});

				notify.success("Less compiled successfully.");
			}
		});

	} catch (err) {

		outputLessError(err, buffer);
		
	}
}


function outputLessError(err, buffer) {

	var lines = buffer.split("\n");
	var line = "";
	var linesLeftToSearch = err.line;
	var file = "";
	var lineInFile = 0;
	var adjacentLinesToShow = 2;
	var i;
	var message = "";

	while (linesLeftToSearch) {
		line = lines[linesLeftToSearch];
		if (line.indexOf(fileDelimiter) > -1 ) {
			file = line.split(fileDelimiter)[1];
			lineInFile = err.line - linesLeftToSearch - 1;
			break;
		}
		linesLeftToSearch--;
	}

	message = "Error: " + err.message +
			"\nFile: " + file +
			"\nLine: " + lineInFile + ", Col: " + err.column +
			"\n\n";

	for (i=0; i<adjacentLinesToShow*2+1; i++)
	{
		message += lines[i+err.line-adjacentLinesToShow] + "\n";
		if (i == adjacentLinesToShow-1) {
			var c = 0;
			while(c < err.column) {
				message += " ";
				c++;
			}
			message += "^\n";
		}
	}

	notify.error(message);
}



function watchFolders(folders,onchange,oncreate,onremove)
{
	if (!folders) return;

	folders.forEach(function(folder,i,a)
	{
		watch.createMonitor(cwd+folder, function(monitor)
		{
			monitor.files[ cwd + folder ];
			monitor.on("changed", onchange);
			monitor.on("created", oncreate);
			monitor.on("removed", onremove);
		});
	});
}


function findFilesWithExtInFolders(ext, folders, callback)
{
	if (!folders) return;

	var filesFound = [];
	var foldersSearched = 0;

	folders.forEach(function(folder,i,a)
	{
		watch.walk(cwd+folder, function(err,files){

			if (err) {
				callback(err);
			}

			for ( var file in files ) {
				if (ext == file.ext)
					filesFound.push(file);
			}

			foldersSearched++;

			if (foldersSearched == folders.length)
				callback(null, filesFound);
		});
	});
}


function buildLessFile(file)
{
	var lessFile = "";
	var viewId = "";
	var lessContent = "";
	var filesToWrite = 0;
	var filesWritten = 0;

	function ucase(){
		return arguments[0].toUpperCase();
	}

	// Make a corresponding less file path
	lessFile = file.replaceExt("less");

	// If a less doesn't exist, create one
	try
	{
		fs.statSync(lessFile);
	}
	catch (e)
	{
		// Using some replace/regex to build a view Id for the CSS.

		// remove the working path
		viewId = file.replace(cwd,"").

		// remove the view path
		replace(viewFolders.filter(function(folder){
			return file.match(new RegExp(folder,"g"));
		})[0],"").

		// remove the extension
		replace("."+viewExt, "").

		// convert path to uppercase
		replace(/\b[a-z]/g, ucase).

		// remove slashes
		replace(/[/]/g, "");

		// Now build a less container with the ID
		lessContent = '// Generated with Lessup' +
				'\n\n#' + viewId +
				'\n{\n' +
				'\t// LessCSS goes here.' +
				'\n}';

		fs.writeFile(lessFile,lessContent,function(err){
			if (err) throw err;
			lessFiles.add(lessFile,lessContent);
			compileLess();
		});
	}
}



module.exports.watch = function(options)
{
	// Defaults
	var settings =
	{
		cwd: null,
		compileTo: "/public/css/style.css",
		lessFolders: ["/views"],
		viewFolders: ["/views"],
		viewExt: "html"
	}

	for ( var option in options ) {
		if (settings.hasOwnProperty(option))
		{
			settings[option] = options[option];
		}
	}

	if (settings.cwd == null) throw "Lessup needs a reference to the app’s current working directory.";

	cwd = settings.cwd;
	compileTo = settings.compileTo;
	lessFolders = settings.lessFolders;
	viewFolders = settings.viewFolders;
	viewExt = settings.viewExt;

	// Merge viewFolders into lessFolders to import any less files that get created
	viewFolders.forEach(function(folder){
		if (lessFolders.indexOf(folder)<0)
			lessFolders.push(folder);
	});

	var lessFilesAdded = [];

	watchFolders(

			lessFolders,

		// on change
			function (f, curr, prev)
			{
				switch (f.ext)
				{
					case viewExt:
						break;

					case "less":
						notify.changed("Less ("+f+") was changed.");
						lessFiles.update(f);
						compileLess();
						break;
				}
			},

		// on create
			function (f, curr, prev)
			{
				switch (f.ext)
				{
					case viewExt:
						try {
							fs.statSync(f.replaceExt("less"));
							notify.added("View ("+f+") was added. Less file exists.");
						} catch (e) {
							lessFilesAdded.push(f.replaceExt("less"));
							notify.added("View ("+f+") was added. Adding Less file.");
							buildLessFile(f);
						}
						break;

					case "less":
						var idx = lessFilesAdded.indexOf(f);
						if ( idx > -1 ) {
							lessFilesAdded = lessFilesAdded.slice(idx,1);
							break;
						}
						notify.added("Less ("+f+") was added.");
						lessFiles.add(f);
						compileLess();
						break;
				}
			},

		// on remove
			function (f, stat)
			{
				switch (f.ext)
				{
					case viewExt:
						notify.removed("View ("+f+") was removed. Leaving Less file in place.");
						break;

					case "less":
						notify.removed("Less ("+f+") was removed.");
						lessFiles.remove(f);
						compileLess();
						break;
				}
			}
	);


	// If view folder is defined, get all views
	if (viewFolders.length)
		findFilesWithExtInFolders(viewExt,viewFolders,function(err,files){

			if (!files) return;

			files.forEach(function(file){
				buildLessFile(file);
			});
		});

	// Get all less files, then compile
	findFilesWithExtInFolders("less",lessFolders,function(err,files){

		if (!files) return;

		files.forEach(function(file){
			lessFiles.add(file);
		});

		compileLess();
	});

};